from .feature_selection_network import FeatureSelectionNetwork, gumbel_sigmoid
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import tqdm
from typing import Literal

def select_gumbel_features(
    X_train,
    y_train,
    device="cpu",
    verbose=False,
    temperature_decay=0.997,
    epochs=300,
    batch_size=1,
    fs_balance=1.0,
    target_features_mode: Literal["auto", "target", "raw"] = "auto",
    mode: Literal["classification", "regression"] = "classification",
):
    if mode == "classification":
        balance = (y_train.sum(axis=0) / len(y_train)).flip(0)
        network = FeatureSelectionNetwork(X_train.shape[1], 32, len(balance)).to(device)
        criterion = nn.CrossEntropyLoss(weight=balance.to(torch.float32).to(device))
    elif mode == "regression":
        network = FeatureSelectionNetwork(X_train.shape[1], 32, 1).to(device)
        criterion = nn.MSELoss()
    else:
        raise ValueError(f"Invalid mode: {mode}")

    trainloader = DataLoader(list(zip(X_train, y_train)), batch_size=batch_size)

    optimizer = optim.Adam(
        [
            {"params": network.fc1.parameters(), "lr": 4e-3},
            {"params": network.cont.parameters(), "lr": 3e-4},
        ]
    )
    temperature = 2.0
    
    if target_features_mode == "auto":
        if X_train.shape[1] > 10000:
            target_features_mode = "raw"
        else:
            target_features_mode = "target"
        
    if target_features_mode == "target":
        target_features = 2
        target_features = target_features / X_train.shape[1]
        
    elif target_features_mode == "raw":
        target_features = None


    for epoch in tqdm.trange(epochs, disable=not verbose):
        total = 0
        correct = 0
        epoch_loss = 0
        for data in trainloader:
            X_, y_ = data
            X_, y_ = X_.to(device), y_.to(device).to(torch.float32)

            optimizer.zero_grad()
            output, selected_no = network(X_.float(), temperature=temperature)
            if mode == "classification":
                proper_y = y_.argmax(axis=1)
            elif mode == "regression":
                proper_y = y_
                
            if target_features is not None:
                loss = criterion(output, proper_y) + fs_balance * (selected_no - target_features) ** 2
            else:
                loss = criterion(output, proper_y) + fs_balance * selected_no
            loss.backward()
            optimizer.step()
            epoch_loss += loss.item()

            predicted = torch.argmax(output, axis=1)
            real = torch.argmax(y_, axis=1)
            total += y_.size(0)
            correct += (predicted == real).sum().item()

            if epoch > epochs:
                temperature = 0
                network.fc1.requires_grad_(False)
                network.emb.requires_grad_(False)
        temperature = temperature * temperature_decay

    features_sum = torch.zeros(X_train.shape[1], dtype=torch.int32)
    for data in trainloader:
        X_, y_ = data
        X_, y_ = X_.to(device), y_.to(device).to(torch.float32)
        features = network.fc1(network.emb)
        features = gumbel_sigmoid(features, tau=0, hard=True)
        features = features.cpu().detach().numpy()
        features_sum += features.sum(axis=0)

    return features_sum.numpy(), network



from .select_gumbel_features import select_gumbel_features
import torch
from typing import Literal


class GFSNetwork:
    def __init__(
        self,
        batch_size=1,
        temperature_decay: float = 0.997,
        epochs: int = 200,
        balance: float = 1.0,
        device: str = "cpu",
        verbose: bool = False,
        mode: Literal["classification", "regression"] = "classification",
    ) -> None:
        """Perform feature selection using GFSNetwork.

        Args:
            batch_size (int, optional): Batch size. Larger batch size speeds up training, but makes feature selection less aggresive. Defaults to 1.
            temperature_decay (float, optional): Temperature decay. Defaults to 0.997.
            epochs (int, optional): Number of epochs. Defaults to 200.
            balance (float, optional): Balance between classification and feature selection. Larger value puts more weight on feature selection. Defaults to 1.0.
            device (str, optional): Device to use. Defaults to "cpu".
            verbose (bool, optional): Verbosity. Defaults to False.
        """
        self.scores_ = None
        self.device = device
        self.verbose = verbose
        self.temperature_decay = temperature_decay
        self.epochs = epochs
        self.network = None
        self.batch_size = batch_size
        self.balance = balance
        self.mode = mode

    def fit(
        self,
        X,
        y,
        scale=True,
        target_features_mode: Literal["auto", "target", "raw"] = "raw",
    ):
        # cast to torch
        if type(X) is not torch.Tensor:
            X = torch.tensor(X)
            y = torch.tensor(y)

        # assert that y is one-hot encoded
        if len(y.shape) == 1 and self.mode == "classification":
            y = torch.nn.functional.one_hot(y.to(int))
        elif len(y.shape) == 1 and self.mode == "regression":
            y = y.view(-1, 1).to(torch.float32)
            
        if scale:
            # perform whitening
            X = (X - X.mean(0)) / (X.std(0) + 1e-6)

        self.scores_, self.network = select_gumbel_features(
            X,
            y,
            self.device,
            self.verbose,
            temperature_decay=self.temperature_decay,
            epochs=self.epochs,
            batch_size=self.batch_size,
            fs_balance=self.balance,
            target_features_mode=target_features_mode,
            mode=self.mode,
        )
        self.ranking_ = self.scores_.argsort()[::-1]
        return self

    @property
    def support_(self):
        assert self.scores_ is not None, "You must call fit before accessing support_"
        return self.scores_ > 0

    def transform(self, X):
        assert self.support_ is not None, "You must call fit before transform"
        return X[:, self.support_]

    def fit_transform(self, X, y):
        self.fit(X, y)
        return self.transform(X)
